import os
import inspect
from pprint import pprint
from builtins import print
from collections import OrderedDict
from sugarbowl import cachedproperty, import_object
from ..settings import settings
from ..task import Task, Prototype, ConvertError, convert
from ..utils import stylize
from .exception import GetattrError


class Module(Task):

    # Public

    meta_auto = True

    def __getattribute__(self, name):
        nested_name = None
        if '.' in name:
            # Nested name - split
            name, nested_name = name.split('.', 1)
        try:
            attribute = super().__getattribute__(name)
        except AttributeError as exception:
            # To get correct AttributeError message here
            if isinstance(exception, GetattrError):
                raise  # pragma: no cover TODO: remove no cover
            raise GetattrError(
                'Module "{self}" has no attribute "{name}".'.
                format(self=self, name=name))
        if nested_name is not None:
            attribute = getattr(attribute, nested_name)
        return attribute

    @property
    def meta_basedir(self):
        default = os.path.abspath(
            os.path.dirname(inspect.getfile(type(self))))
        return self.meta_inspect('basedir', default=default)

    @property
    def meta_cache(self):
        """Vars caching status (enabled or disabled).
        """
        return self.meta_inspect('cache', default=settings.cache)

    @classmethod
    def meta_create(cls, *args, **kwargs):
        # Create module object
        spawned_class = cls.meta_spawn()
        self = super(Module, spawned_class).meta_create(*args, **kwargs)
        # Create tasks
        names = []
        for cls in type(self).mro():
            for name, attr in vars(cls).items():
                if name in names:
                    continue
                names.append(name)
                if isinstance(attr, Prototype):
                    task = attr.meta_build(meta_module=self)
                    setattr(type(self), name, task)
        return self

    @property
    def meta_default(self):
        return self.meta_inspect('default', default='list')

    def meta_invoke(self, *args, **kwargs):
        default = getattr(self, self.meta_default)
        result = default(*args, **kwargs)
        return result

    @cachedproperty
    def meta_listeners(self):
        listeners = self.meta_inspect('listeners')
        if listeners is None:
            listeners = []
            if not self.meta_module:
                for pointer in settings.listeners:
                    element = import_object(pointer)
                    listener = element()
                    listeners.append(listener)
        return listeners

    def meta_inspect(self, name, *, module=False, default=None):
        """Return meta parameter
        """
        return super().meta_inspect(name, default=default)

    @classmethod
    def meta_spawn(cls):
        names = []
        attrs = {}
        for namespace in cls.mro():
            for name, attr in vars(namespace).items():
                if name in names:
                    continue
                names.append(name)
                if name.isupper():
                    continue
                elif name.startswith('_'):
                    continue
                elif name.startswith('meta_'):
                    continue
                elif isinstance(attr, Prototype):
                    attrs[name] = attr.meta_fork()
                else:
                    if cls.meta_auto:
                        try:
                            attrs[name] = convert(attr)
                        except ConvertError:
                            pass
        attrs['__doc__'] = cls.__doc__
        attrs['__module__'] = cls.__module__
        return type(cls)(cls.__name__, (cls,), attrs)

    @property
    def meta_style(self):
        return self.meta_inspect('style', default='module')

    @property
    def meta_tasks(self):
        """Module's tasks dict-like object.

        Dict contains task instances, not values.
        """
        tasks = {}
        for name, attr in vars(type(self)).items():
            if isinstance(attr, Task):
                tasks[name] = attr
        return tasks

    def meta_update(self):
        for task in self.meta_tasks.values():
            task.meta_update()
        super().meta_update()

    def list(self, module=None):
        """Print module tasks.
        """
        # TODO: exception here breaks system tests. Why???
        # Example:
        # raise Exception()
        names = []
        module = self.__get_task(module)
        for name in sorted(dir(module)):
            # TODO: code duplication with ModuleMetaclass.meta_spawn
            if name.isupper():
                continue  # pragma: no cover TODO: remove no cover
            elif name.startswith('_'):
                continue
            elif name.startswith('meta_'):
                continue
            elif name in module.meta_tasks:
                task = module.meta_tasks[name]
                if task.meta_hidden:
                    continue
                name = stylize(task.meta_qualname, style=task.meta_style)
            else:
                name = '.'.join(filter(None, [module.meta_qualname, name]))
            names.append(name)
        result = '\n'.join(names)
        print(result)

    def info(self, task=None):
        """Print task information.
        """
        info = ''
        task = self.__get_task(task)
        info += task.meta_qualname
        info += task.meta_signature
        info += '\n---\n'
        info += 'Type: ' + task.meta_type
        info += '\n'
        info += 'Dependencies: ' + str(task.meta_dependencies)
        info += '\n'
        info += 'Default arguments: ' + str(task.meta_args)
        info += '\n'
        info += 'Default keyword arguments: ' + str(task.meta_kwargs)
        info += '\n---\n'
        info += task.meta_docstring
        print(info)

    def meta(self, task=None):
        """Print task metadata.
        """
        meta = OrderedDict()
        task = self.__get_task(task)
        for name in sorted(dir(task)):
            if name.startswith('meta_'):
                key = name.replace('meta_', '')
                attr = getattr(task, name)
                if not inspect.ismethod(attr):
                    meta[key] = attr
        pprint(meta)

    # Private

    # TODO: somehow to merge with __getattribute__?
    def __get_task(self, name=None):
        if name is None:
            return self
        nested_name = None
        if '.' in name:
            # Nested name - split
            name, nested_name = name.split('.', 1)
        # TODO: add good exception text here like in __getattribute__
        task = self.meta_tasks[name]
        if nested_name is not None:
            task = task.__get_task(nested_name)
        return task
